﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.Linq;
using System.Reactive.Disposables;
using System.Reactive.Linq;
using System.Reactive.Subjects;
using System.Threading.Tasks;
using System.Windows.Input;
using Windows.ApplicationModel;
using Windows.ApplicationModel.Core;
using Windows.Foundation;
using Windows.Globalization;
using Windows.Media.SpeechRecognition;
using Windows.UI.Core;
using KakiAISpeaker.Client.Helpers;
using KakiAISpeaker.Client.Models;
using Microsoft.Bot.Connector.DirectLine;
using Newtonsoft.Json;
using Reactive.Bindings.Extensions;
using WebSocketSharp;

namespace KakiAISpeaker.Client.ViewModels
{
    /// <summary>
    ///     메인 페이지 뷰모델
    /// </summary>
    public class MainViewModel : BindableBase
    {
        /// <summary>
        ///     다이렉트 라인 시크릿
        /// </summary>
        private const string DIRECT_LINE_SECRET = "PZ0mGcox1m4.cwA.H-I.GGuo1qzJ6KCfE7QpmjNDw2EPk02xkhOlGKsd912iI48";

        private const string SRGS_FILE_NAME = "SRGS.xml";

        private readonly string _botId = "KakiAISpeakerChannel";

        private readonly CompositeDisposable _compositeDisposable = new CompositeDisposable();
        private readonly string _fromUser = "kaki104";

        /// <summary>
        ///     클라이언트 상태를 주제로 하는 서브젝트
        /// </summary>
        private readonly Subject<ClientState> _subject = new Subject<ClientState>();

        /// <summary>
        ///     봇클라이언트
        /// </summary>
        private DirectLineClient _botClient;

        /// <summary>
        ///     컨버세이션
        /// </summary>
        private Conversation _conversation;

        private ClientStates _currentClientState;

        private string _inputText;
        private bool _isConversation;
        private int _retryCount;
        private SpeechRecognizer _speechRecognizer;
        private string _watermark;

        private WebSocket _webSocketClient;

        /// <summary>
        ///     생성자
        /// </summary>
        public MainViewModel()
        {
            if (!DesignMode.DesignModeEnabled) Init();
        }

        /// <summary>
        ///     전송 버튼 커맨드
        /// </summary>
        public ICommand SendCommand { get; set; }

        /// <summary>
        ///     다이얼로그 목록
        /// </summary>
        public IList<Activity> Dialogs { get; set; }

        /// <summary>
        ///     입력받은 텍스트
        /// </summary>
        public string InputText
        {
            get => _inputText;
            set => Set(ref _inputText, value);
        }

        public ClientStates CurrentClientState
        {
            get => _currentClientState;
            set => Set(ref _currentClientState, value);
        }

        /// <summary>
        ///     초기화
        /// </summary>
        private async void Init()
        {
            Dialogs = new ObservableCollection<Activity>();
            SendCommand = new RelayCommand(ExecuteSendCommand);

            //봇 컨버세이션 시작
            await StartBotConversationAsync();

            //음성 인식 초기화 - 한글을 지원하지 않기 때문에 영문으로 인식하도록 함
            var supportedLanguages = SpeechRecognizer.SupportedGrammarLanguages;
            var enUS = supportedLanguages.FirstOrDefault(p => p.LanguageTag == "en-US")
                       ?? SpeechRecognizer.SystemSpeechLanguage;
            await InitializeRecognizerAsync(enUS);
        }

        private async Task InitializeRecognizerAsync(Language language)
        {
            try
            {
                var grammarContentFile = await Package.Current.InstalledLocation.GetFileAsync(SRGS_FILE_NAME);
                if (grammarContentFile == null)
                    throw new NullReferenceException("SRGS 파일이 존재하지 않습니다.");

                // Create an instance of SpeechRecognizer.
                _speechRecognizer = new SpeechRecognizer(language);

                //윈도우 지원 음성 인식
                var resultGeneratedObservable =
                    Observable
                        .FromEvent<TypedEventHandler<SpeechContinuousRecognitionSession,
                                SpeechContinuousRecognitionResultGeneratedEventArgs>,
                            SpeechContinuousRecognitionResultGeneratedEventArgs>(
                            h => (sender, args) => h(args),
                            h => _speechRecognizer.ContinuousRecognitionSession.ResultGenerated += h,
                            h => _speechRecognizer.ContinuousRecognitionSession.ResultGenerated -= h);

                resultGeneratedObservable
                    .Where(args => !(args.Result.Confidence == SpeechRecognitionConfidence.Low
                                     || args.Result.Confidence == SpeechRecognitionConfidence.Rejected)
                                   && args.Result.SemanticInterpretation.Properties.Any(p => p.Key == "ACTION"))
                    .Select(args => new ClientState {States = ClientStates.StartConversation, Data = args})
                    .Subscribe(_subject)
                    .AddTo(_compositeDisposable);

                _subject
                    .Subscribe(async state =>
                    {
                        //CurrentClientState = state.States;
                        Debug.WriteLine(state.States);
                        switch (state.States)
                        {
                            case ClientStates.Idle:
                                Debug.WriteLine($"Idle {DateTime.Now:O}");
                                break;
                            case ClientStates.StartConversation:
                                var args = (SpeechContinuousRecognitionResultGeneratedEventArgs) state.Data;
                                var action =
                                    args.Result.SemanticInterpretation.Properties
                                        .FirstOrDefault(p => p.Key == "ACTION");
                                if (string.IsNullOrEmpty(action.Key)) return;

                                var command = action.Value.FirstOrDefault();
                                switch (command)
                                {
                                    case "BEGIN":
                                        //시스템 메시지 출력
                                        await ExecuteDispatcherRunAsync(async () =>
                                        {
                                            Dialogs.Add(new Activity
                                            {
                                                Text = $"[System Message] Start Conversation"
                                            });
                                            await SendMessageToBotAsync("Start");
                                            //PlaySystemMessage("hello.mp3");
                                        });
                                        _isConversation = true;
                                        _retryCount = 0;
                                        break;
                                }

                                break;
                            case ClientStates.PlaySystemVoice:
                                break;
                            case ClientStates.StopSystemVoice:
                                break;
                            case ClientStates.StartRecoding:
                                break;
                            case ClientStates.StopRecoding:
                                break;
                            case ClientStates.SendVoiceCommand:
                                break;
                            case ClientStates.ReceiveVoiceCommandResult:
                                break;
                            case ClientStates.PlayVoiceCommandResult:
                                break;
                            case ClientStates.StopVoiceCommandResult:
                                break;
                            case ClientStates.EndConversation:
                                break;
                            default:
                                throw new ArgumentOutOfRangeException();
                        }
                    });

                //SRGS 읽어서 조건에 추가
                var grammarConstraint = new SpeechRecognitionGrammarFileConstraint(grammarContentFile);
                _speechRecognizer.Constraints.Add(grammarConstraint);

                // Compile the constraint.
                var compilationResult = await _speechRecognizer.CompileConstraintsAsync();
                // Check to make sure that the constraints were in a proper format and the recognizer was able to compile it.
                if (compilationResult.Status != SpeechRecognitionResultStatus.Success)
                    Debug.WriteLine("Error SpeechRecognizer.CompileConstraints");

                //연속 음성 인식 시작 - 시작 단어 확인하기 위해
                await _speechRecognizer.ContinuousRecognitionSession.StartAsync();
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex);
                throw;
            }
        }

        /// <summary>
        ///     봇 초기화
        /// </summary>
        /// <returns></returns>
        private async Task StartBotConversationAsync()
        {
            //다이렉트라인 클라이언트 생성
            _botClient = new DirectLineClient(DIRECT_LINE_SECRET);
            _conversation = await _botClient.Conversations.StartConversationAsync();

            Debug.WriteLine($"ConversationId : {_conversation.ConversationId}");

            //웹소켓 클라이언트
            _webSocketClient = new WebSocket(_conversation.StreamUrl);
            _webSocketClient.OnMessage += WebSocketClient_OnMessage;
            _webSocketClient.Connect();
        }

        /// <summary>
        ///     소켓 메시지 수신
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private async void WebSocketClient_OnMessage(object sender, MessageEventArgs e)
        {
            if (string.IsNullOrWhiteSpace(e.Data))
                return;

            var activitySet = JsonConvert.DeserializeObject<ActivitySet>(e.Data);
            //언젠가 쓰일지도 모르는..
            _watermark = activitySet.Watermark;

            //botId와 맞는 녀석만 골라내고..
            var activities = from x in activitySet.Activities
                where x.From.Id == _botId
                select x;

            foreach (var activity in activities) await ExecuteDispatcherRunAsync(() => Dialogs.Add(activity));
        }

        /// <summary>
        ///     전송 버튼 커맨드 실행
        /// </summary>
        private async void ExecuteSendCommand()
        {
            if (string.IsNullOrEmpty(InputText)) return;
            await SendMessageToBotAsync(InputText);
            InputText = string.Empty;
        }

        /// <summary>
        ///     메시지 송신
        /// </summary>
        /// <param name="inputText"></param>
        /// <returns></returns>
        private async Task SendMessageToBotAsync(string inputText)
        {
            var userMessage = new Activity
            {
                From = new ChannelAccount(_fromUser),
                Text = inputText,
                Type = ActivityTypes.Message
            };

            Dialogs.Add(userMessage);

            await _botClient.Conversations.PostActivityAsync(_conversation.ConversationId, userMessage);
        }

        /// <summary>
        ///     디스패처 실행
        /// </summary>
        /// <param name="handler"></param>
        /// <returns></returns>
        private async Task ExecuteDispatcherRunAsync(DispatchedHandler handler)
        {
            await CoreApplication.MainView.Dispatcher
                .RunAsync(CoreDispatcherPriority.Normal, handler);
        }
    }
}
